/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Dict
 *
 * The Initial Developer of the Original Code is David Costanzo
 * Portions created by the Initial Developer are Copyright (C) 2002-2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   David Costanzo <david_costanzo at yahoo dot com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// TODO: Don't pollute the global namespace
const DICT_STATE_Start                = 0;
const DICT_STATE_WantClientAck        = 1;
const DICT_STATE_WantDefinition       = 2;
const DICT_STATE_WantDefinitionsDone  = 3;
const DICT_STATE_WantStrategiesInfo   = 4;
const DICT_STATE_WantStrategies       = 5;
const DICT_STATE_WantMatchesInfo      = 6;
const DICT_STATE_WantMatches          = 7;
const DICT_STATE_Done                 = 8;
const DICT_STATE_Error                = -1;

var dictStringBundle;

function Observer(connection, word) {
    this._connection = connection;
    this._word       = word;

    this._bufferedData = '';
    this._state = DICT_STATE_Start;

    this._definitions  = [];
}

// Creates a new window and sets the document to be the given HTML.
// Resizes the window to fit the content.
Observer.prototype.createWindow = function (htmlDocument)
{

    const features = [
        'status=no', 
        'toolbar=no',
        'menubar=no',
        'location=no',
        'width=450',
        'height=100'
    ];

    var popupWindow = window.open(
        '',
        '',
        features.join(','));
    
    popupWindow.document.clear();
    popupWindow.document.write(htmlDocument);
    popupWindow.document.close();
    
    // HACK: Figure out how to size the window
    //       to fit the content
    popupWindow.sizeToContent();
    popupWindow.innerHeight += 70;

    // If it's too large to display without scrollbars, shorten it
    if (600 < popupWindow.innerHeight) {
        // limit the height
        popupWindow.innerHeight = 600;
        
        // make wider to account for the width of the 
        // vertical scrollbar.  If we don't do this, a 
        // horizontal scrollbar will appear.
        popupWindow.innerWidth += 20;  
    }

    return popupWindow;
}


// wraps some HTML inside HTML/BODY tag and adds a title
Observer.prototype.createHtmlDocument = function (title, body)
{
    var htmlDocument;
    
    htmlDocument = 
        '<HTML>\n'    +
        '  <HEAD><TITLE>' + title + '</TITLE></HEAD>\n' +
        '  <BODY>\n'  +
        body          +
        '  </BODY>\n' +
        '</HTML>\n';

    return htmlDocument;
}

// Reports that there was an protocol error.
// This should only happen if the server doesn't obey the protocol
// (or if there are bugs in the client state machine)
Observer.prototype.reportStateMachineError = function ()
{
    throw 
        'Unknown problem encountered when defining "' + this._word + '"\n' +
        'If problem presists, please file a bug report on the Dict project\n';
}

// Pops a DICT response line from the _bufferedInput member variable.
// On success, returns true and sets this._responseLine, this._responseCode, 
// and this._responseLineBody.
// Return false, if there wasn't a complete response line.
Observer.prototype.popResponseLine = function ()
{
    var responseLinePattern = /^(\d+)\s*(.*)\n/;
    
    var result = this._bufferedData.match(responseLinePattern);
    if (result == null) {
        return false;
    }

    this._responseLine     = result[0];
    this._responseCode     = result[1];
    this._responseLineBody = result[2];
    
    // remove the response line from the buffered input
    var lineEndIndex   = this._bufferedData.indexOf('\n');
    this._bufferedData = this._bufferedData.slice(lineEndIndex + 1);
    return true;
}

// Performs a graceful shutdown
Observer.prototype.gracefulShutdown = function ()
{
    this._connection.sendData('QUIT\r\n');
    this._bufferedData = '';
    this._state = DICT_STATE_Done;
    
    this._connection.disconnect();
}


// Parses and returns all buffered input up to (but not including) 
// the sequence CRLF . CRLF.  The multiline input is then removed 
// from the buffered input.
// Returns null if the multiline input is not complete.
//
// The DICT protocol uses multiline input of this form to return 
// strategies, matches, and definitions. 
Observer.prototype.popMultilineInput = function () 
{
    // NOTE: We should like for \r\n.\r\n, but all "\r" are removed
    //       from the input stream to make regexp parsing easier.
    const sentinal = "\n.\n";

    // Check to see if we have the entire list
    var endIndex = this._bufferedData.indexOf(sentinal);
    if (endIndex == -1) {
        return null;  // the multiline input isn't complete
    }
    
    // pop off and return the multiline input
    var multilineData = this._bufferedData.slice(0, endIndex);
    
    // remove the multiline input and sentinal sequence from the buffered input
    this._bufferedData = this._bufferedData.slice(endIndex + sentinal.length);

    return multilineData;
}

Observer.prototype.onStreamDataAvailable = function (
    request, 
    inStream, 
    offset, 
    count
) 
{
    try {
        // read the data from the socket into our buffer
        var newData = this._connection.readData(0, count);

        // strip out carriage returns, since they complicate regexp usage
        this._bufferedData += newData.replace(/\r/g, '');

        // do {
        //   {handle data}
        // } while( {there's enough data to handle} );
        do {
            var previousLength = this._bufferedData.length;

            switch (this._state) {

            case DICT_STATE_Start:
                if (this.popResponseLine()) {
                    
                    if (this._responseCode == 220) {
                        // 220 <text> <capabilities> <msg-id> CRLF
                        // (this is the session init sequence)
                        this._connection.sendData('CLIENT Mozilla Dict plugin\r\n');
                        this._connection.sendData('DEFINE * "' + this._word + '"\r\n');
                    
                        this._state = DICT_STATE_WantClientAck;
                    }
                    else {
                        this.reportStateMachineError();
                    }
                }
                break;


            case DICT_STATE_WantClientAck:

                if (this.popResponseLine()) {
                    if (this._responseCode == 250) {
                        this._state = DICT_STATE_WantDefinitionsDone;
                    }
                    else {
                        this.reportStateMachineError();
                    }
                }
                break;

            case DICT_STATE_WantDefinitionsDone:

                if (this.popResponseLine()) {
                    if (this._responseCode == 150) {
                        // 150: n definitions returned
                    }
                    else if (this._responseCode == 151) {
                        // 151: <word> <database> <name>
                        // The definition follows this line
                        this._state = DICT_STATE_WantDefinition;
                    }
                    else if (this._responseCode == 250) {

                        // We have all definitions

                        // Format the definitions into HTML
                        var definitionDocumentBody = '<TABLE width="450"><TR><TD>';

                        var i;
                        for (i = 0; i < this._definitions.length; i++) {

                            definitionDocumentBody += '<B>';
                            definitionDocumentBody += this._definitions[i].source;
                            definitionDocumentBody += '</B><BR>\n';
                        
                            var deftext = this._definitions[i].text;

                            // Escape HTML entities
                            deftext = deftext.replace(/&/g,   '&amp;');
                            deftext = deftext.replace(/</g,   '&lt;');
                            
                            // replace newlines with <BR> and whitespace with
                            // non-breaking spaces to preserve formatting.
                            // This is better than using <PRE> because it
                            // allows us to use other mark-up tags.
                            deftext = deftext.replace(/\n/g,  '<BR>');
                            deftext = deftext.replace(/ /g,   '&nbsp;');
                            deftext = deftext.replace(/\t/g,  '&nbsp;&nbsp');

                            // Words that are cross-referenced by WordNet show up 
                            // in braces. e.g., "{word}" and "{word(P)}"
                            deftext = deftext.replace(/{(.*?)(\(.*\))?}/g, '$1');

                            // add newlines after <BR> for readability
                            deftext = deftext.replace(/<BR>/g, '<BR>\n');

                            definitionDocumentBody += '<TT>'+deftext+'</TT><BR>\n';

                            // put a visual divider between this entry and the next
                            if (i + 1 < this._definitions.length) {
                                definitionDocumentBody += '<BR>\n';
                            }
                        }

                        var definitionDocument = this.createHtmlDocument(
                            this._word,
                            definitionDocumentBody);

                        this.createWindow(definitionDocument);

                        this.gracefulShutdown();
                    }
                    else if (this._responseCode == 550) {
                        // 550 invalid database

                        // TODO: Handle this gracefully
                        this.reportStateMachineError();
                    }
                    else if (this._responseCode == 552) {
                        // 552 no match

                        // Present the user with a nearest match 
                        this._connection.sendData('SHOW STRAT\r\n');
                        this._state = DICT_STATE_WantStrategiesInfo;
                    }
                    else {
                        this.reportStateMachineError();
                    }
                }
                break;

            case DICT_STATE_WantDefinition:
                
                // 151 <word> <database> CRLF
                // <definition line 1>   CRLF
                // <definition line 2>   CRLF
                // . . . 
                // <definition line n>   CRLF
                // . CRLF
                
                // determine if we have the entire definition
                var definitionText = this.popMultilineInput();
                if (definitionText == null) {
                    break; // wait for the rest of the definition
                }

                var definition = new Object;

                // 151 <word> <database> CRLF
                var result = this._responseLineBody.match(/^(\S+)\s+\S+\s+(.*)$/);
                if (result != null) {
                    definition.word   = result[1];
                    definition.source = result[2];
                }

                // Store the definition
                definition.text = definitionText;

                this._definitions.push(definition);

                this._state = DICT_STATE_WantDefinitionsDone;
                break;

            case DICT_STATE_WantStrategiesInfo:
                if (this.popResponseLine()) {
                    if (this._responseCode == 111) {
                        // 111 n stragegies available
                        this._state = DICT_STATE_WantStrategies;
                    }
                    else if (this._responseCode == 555) {
                        // 555 No strategies available
                        this.gracefulShutdown();
                        alert('No definitions found for "' + this._word + '"');
                    }
                    else {
                        // unexpected response
                        this.reportStateMachineError();
                    }
                }
                break;

            case DICT_STATE_WantStrategies:

                // <strategy 1> <description 1> CRLF
                // <strategy 2> <description 2> CRLF
                // . . . 
                // <strategy n> <description n> CRLF
                // . CRLF
                
                // determine if we have the entire strategy list
                var strategies = this.popMultilineInput();
                if (strategies == null) {
                    break; // wait for the rest of the strategies
                }


                // search for the "Levenshtein distance one"
                var haveLevStrategy = false;
                // <strategy> <description> CRLF
                var matchExp = /^(\S+)\s+(.*)$/gm;
                while ((result = matchExp.exec(strategies)) != null) {
                    var strategy    = result[1];
                    var description = result[2];

                    if (strategy == 'lev') {
                        haveLevStrategy = true;
                    }
                }

                if (haveLevStrategy) {
                    // Use the "lev" strategy to find nearest matches
                    this._connection.sendData('MATCH * lev "' + this._word + '"\r\n');
                    this._state = DICT_STATE_WantMatchesInfo;
                }
                else {
                    // No good matching strategies
                    this.gracefulShutdown();
                    
                    alert('No definitions found for "' + this._word + '"');
                }

                break;

            case DICT_STATE_WantMatchesInfo:

                if (this.popResponseLine()) {
                    if (this._responseCode == 250) {
                        // 250 ok
                        // This is from the MATCH request.  Ignore it.
                    }
                    else if (this._responseCode == 152) {
                        // 152 n matches found
                        this._state = DICT_STATE_WantMatches;
                    }
                    else if (this._responseCode == 552) {
                        // 552 no match
                        this.gracefulShutdown();
                        alert('No definitions found for "' + this._word + '"');
                    }
                    else {
                        // unexpected response
                        this.reportStateMachineError();
                    }
                }
                break;

            case DICT_STATE_WantMatches:

                // <database 1> <word 1> CRLF
                // <database 2> <word 2> CRLF
                // . . . 
                // <database n> <word n> CRLF
                // . CRLF
                
                // determine if we have the entire strategy list
                var matches = this.popMultilineInput();
                if (matches == null) {
                    break; // wait for the rest of the matches
                }

                // <database> <word> CRLF
                var words    = new Object;
                var matchExp = /^(\S+)\s+(.*)$/gm;
                while ((result = matchExp.exec(matches)) != null) {
                    var database    = result[1];
                    var word        = result[2];

                    var normalizedWord = word.toLowerCase();

                    words[normalizedWord] = true;
                }

                // REVISIT: There has got to be a more direct way
                //   of counting the number of elements in an associative
                //   array.
                var totalMatches = 0;
                for (word in words) {
                    totalMatches++;
                }

                if (totalMatches == 0) {
                    // There are no matches.

                    // This is probably a bug in either the server or the
                    // client, since we should not have transitioned to
                    // the WantMatches state if there were no matches.
                    this.gracefulShutdown();

                    alert('No definitions found for "' + this._word + '"');
                }
                else if (totalMatches == 1) {
                    // There is exactly one match.
                    // Rather than present just one match, we can save the
                    // user some trouble and just request the definition
                    // of the match.

                    var word;
                    for (word in words) {
                        // Request a defintion for word.
                        this._connection.sendData('DEFINE * ' + word + '\r\n');
                        this._state = DICT_STATE_WantClientAck;
                    }
                }
                else {
                    // There is more than one closest match
                    // Display them all and let the user decide which to define
                    var matchesDocumentBody = '<B>Closest Matches</B><BR>\n';

                    var word;
                    for (word in words) {
                        matchesDocumentBody += word + '<BR>\n';
                    }

                    var matchesDocument = this.createHtmlDocument(
                        this._word + ' - Closest Matches',
                        matchesDocumentBody);

                    this.createWindow(matchesDocument);
                    
                    this.gracefulShutdown();
                }

                break;

            case DICT_STATE_Done:
                break;

            default:
                reportStateMachineError();
                break;
            }
            
        } while (previousLength != this._bufferedData.length);
    }
    catch (ex) {
        this._connection.disconnect();
        alert(ex);
    }
}

Observer.prototype.onStreamClose = function (
    status
) 
{
}

        
function DefineWord(host, port, word)
{
    try {
        var connection = new CBSConnection();
        
        connection.connect(host, port, (void 0), true, null);
        
        var observer = new Observer(connection, word);
        connection.startAsyncRead(observer);
    }
    catch (ex) {
        alert(ex);
    }
}

function GetSelection() 
{
    var selectedWord = gContextMenu.searchSelected();

    // sanitize the selection
    selectedWord = selectedWord.replace(/^\W+/, ''); // trim leading non-alnum chars
    selectedWord = selectedWord.replace(/\W+$/, ''); // trim trailing non-alnum chars
    selectedWord = selectedWord.replace(/\s+/, ' '); // conflate whitespace runs

    return selectedWord;
}

        
function DefineSelection()
{
    var host = 'dict.org';
    var port = 2628;

    var selectedWord = GetSelection();

    if (selectedWord.indexOf('"') != -1) {
        alert("Sorry, selections that contain quotation marks cannot be defined.");
        return;
    }

    if (selectedWord.length == 0) {
        alert("Please select a word to define.");
        return;
    }

    DefineWord(host, port, selectedWord);
}


// Every time a new browser window is made dictMenuInit will be called
window.addEventListener("load", dictInit, false);

function dictInit() {

    dictStringBundle = document.getElementById("dict_string_bundle");

    var contextMenu = document.getElementById("contentAreaContextMenu");

    // The first time dictInit is called, contextMenu is null.
    // This test prevents an error from being logged in the JavaScript Console
    if (contextMenu != null) {
        contextMenu.addEventListener(
            "popupshowing",
            onDictPopup,
            false);  
    }
}


// hide the Dictionary Search item when apropriate (use same logic as for web search)
function onDictPopup() {

    var item = document.getElementById('dict-define-selection');
    if (gContextMenu.isTextSelected) {

        var selection = GetSelection();
        
        const LENGTHCUTOFF = 15;

        if (LENGTHCUTOFF < selection.length) {
            // selection is too long to appear in menu.  It must be abbreviated.
            selection = selection.slice(0, LENGTHCUTOFF) + '...';
        }

        var labelTemplate = dictStringBundle.getString("context_searchselect.label");

        item.setAttribute("label", labelTemplate.replace('%S', selection));
        item.hidden = false; // show the menu item
    }
    else {
        item.hidden = true; // hide the menu item
    }
}
